package db;

# This file is part of sqlsus
#
# Copyright (c) 2008,2009 sativouf
# http://sqlsus.sf.net/
#
# Licensed under the GPL version 2.0 license.
# See LICENSE file or
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html

use strict;
use warnings;

sub clone_table {
	my $table = shift;
	my $columns = shift;

	last if $main::interrupt;
	
	my @cols;
	if ($columns eq "*") {
		@cols = &db::query("SELECT column_name FROM databases WHERE database_name = ? AND table_name = ?", $main::database, $table);
		if (not @cols) {
			print STDERR "Error: Impossible to determine the columns for \"$table\"... Please \"get columns $table\" first, or specify columns.\n";
			return;
		} else {
			$columns = join ',', @cols;
		}
	}
	my $clonedb = "$main::dumpdbbase.$main::database.db";
	my $dbhout = DBI->connect("dbi:SQLite:dbname=$clonedb","","", {AutoCommit => 0});
	$dbhout->do("CREATE TABLE IF NOT EXISTS $table ($columns)");
	my ($rows_already_fetched) = $dbhout->selectrow_array("SELECT count(*) FROM $table");

	if ($rows_already_fetched) {
		print STDERR "[+] cloning $main::database.$table (resuming at row $rows_already_fetched)\n";
	} else {
		print STDERR "[+] cloning $main::database.$table\n";
	}

	my ($hits, $log_table) = &select::mass_query("SELECT $columns FROM $table LIMIT $rows_already_fetched,99999999");
	if (not $log_table) {
		$dbhout->disconnect;
		return 0;
	}

	my $dbhin = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");

	my ($limit) = $dbhin->selectrow_array("SELECT count(*) FROM $log_table");

	my @header = &get_header($log_table);

	if ($main::interrupt and $conf::blind) {
		print STDERR "Interrupted while cloning in blind mode, only storing the first completed rows\n";
		my @real_columns = map { $_ = &get_real_column($log_table,$_) } @header;
		my $i;
		for $i (0..$limit) {
			my ($row_has_null) = &query("SELECT count(*) FROM $log_table WHERE row = $i AND (" . join (' IS NULL OR ', @real_columns) . " IS NULL)");
			if ($row_has_null) {
				$limit = $i;
				last;
			}
		}
	}

	# remove some unwanted chars
	for (@header) {
		s/[\\'"\x00\(\)]//g;
	}
	my $sthin = $dbhin->prepare("SELECT * FROM $log_table ORDER BY ABS(row) LIMIT $limit");
	$sthin->execute();

	while (my @row = $sthin->fetchrow_array) {
		my @placeholders;
		my @values;

		shift @row;
		for my $value (@row) {
			if (defined $value and $value eq "\x06") {
				push @placeholders, "NULL";
			} else {
				push @placeholders, "?";
				push @values, $value;
			}
		}
		$dbhout->do("INSERT INTO $table VALUES(" . join(',', @placeholders) . ")", undef, @values);
	}

	$dbhout->commit();
	&drop_table($log_table);
}


##################################
sub clone {
	my $args = shift;

	my $table;
	my $columns = "*";

	if (not $main::database) {
		print STDERR "Error : current database not defined, use \"start\" first\n";
		return 0;
	}

	if ($args) {
		if ($args =~ /^(\S+)(\s+.*)?$/) {
			$table = $1;
			$columns = $2 if $2;
			$columns =~ s/\s+//;
		}
	}
	
	my @tables;
	if ($table) {
		push @tables, $table;
	} else {
		@tables = &db::query("SELECT DISTINCT table_name FROM databases WHERE database_name = ?", $main::database);
	}

	if (@tables) {
		for my $table (@tables) {
			last if $main::interrupt;
			&clone_table($table, $columns);
		}
	} else {
		print STDERR "No tables nor columns stored for this session. Try \"get db\" first, or specify the table and the columns to clone.\n";
	}
}

##################################
sub init_databases {
	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	$dbh->do("CREATE TABLE IF NOT EXISTS queries ('query', 'table_name')");
	$dbh->do("CREATE TABLE IF NOT EXISTS tables ('table_name', 'column_name')");
	$dbh->do("CREATE TABLE IF NOT EXISTS databases ('database_name', 'table_name', 'column_name', 'count')");
	$dbh->do("CREATE TABLE IF NOT EXISTS privileges ('grantee', 'privilege_type')");
	$dbh->do("CREATE TABLE IF NOT EXISTS history ('command')");
}


##################################
# create temporary sqlite table
# return undef if creation failed.
sub create_table {
	my $query = shift;
	my @header = @_;

	# time to order the queries
	# $$ to distinguish sessions
	my $temp_table = "query_" . time . "_${$}_" . &functions::generate_random_string(10, 1..9);

	# allow to keep weird names, with quotes, NULL, etc
	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","",{AutoCommit => 0});

	$dbh->do("INSERT INTO queries VALUES (?,?)", undef, ($query, $temp_table));

	my $sth = $dbh->prepare("INSERT INTO tables VALUES (?,?)");
	for (@header) {
		$sth->execute(($temp_table, $_));
	}
	$sth->finish();

	$dbh->commit();

	$dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");

	my @nums = (0..1000);
	$dbh->do("CREATE TABLE $temp_table (row, x" . join (',x', @nums[0..$#header]) . ")");
	$dbh->disconnect;
	
	return $temp_table;
}

##################################
sub get_header {
	my $table = shift;
	
	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	my @result = map { $_ = $_->[0] } @{ $dbh->selectall_arrayref("SELECT column_name FROM tables WHERE table_name = ?", undef, $table) };

	return @result;
}


##################################
sub insert_log {
	my ($table, $row, $items, @contents) = @_;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","", {AutoCommit => 0});
	
	my $sql = "INSERT INTO $table VALUES (?, " . join(',', ('?') x $items) . ")";
	my $sth = $dbh->prepare($sql);

	for (my $i = 0; $i < $#contents + 1; $i += $items) {
		$sth->execute($row++, @contents[$i..$i+$items-1]);
	}
	$sth->finish();
	$dbh->commit();
}

##################################
sub insert_item {
	my ($table, $row, $index, $result) = @_;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	my @res = $dbh->selectrow_array("SELECT row FROM $table WHERE row = $row");
	if (@res) {
		$dbh->do("UPDATE $table SET x$index = ? WHERE row = $row", undef, $result);
	} else {
		my @columns = map { $_ = "NULL" } &get_header($table);
		$columns[$index] = "?";
		$dbh->do("INSERT INTO $table VALUES ($row," . join (',', @columns) . ")", undef, $result);
	}
}

##################################
#- remove the internal row column and return an array
sub table_to_array {
	my $table = shift;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	my @columns = map { $_ = &get_real_column($table, $_) } &get_header($table);
	
	my $sth = $dbh->prepare("SELECT row," . join(',', @columns) . " FROM $table ORDER BY ABS(row)");
	$sth->execute();
	
	my @result;
    while (my @values = $sth->fetchrow_array()) {
		# remove row
		shift @values;
        push @result, @values;
    }
    return @result;
}

##################################
sub get_real_column {
	my $table = shift;
	my $column = shift;

	my @columns = &get_header($table);
	my( $i ) = grep { $columns[$_] eq $column } 0 .. $#columns;
	my @nums = (0..10000);
	return "x$nums[$i]";
}

	
##################################
sub drop_table {
	my $table = shift;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","", {AutoCommit => 0});
	$dbh->do("DELETE FROM tables WHERE table_name = '$table'");
	$dbh->do("DELETE FROM queries WHERE table_name = '$table'");
	$dbh->do("DROP TABLE $table");
	$dbh->commit();
}

##################################
sub query_in_log {
	my $query = shift;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	return $dbh->selectrow_array("SELECT table_name FROM queries WHERE query = ?", undef, $query);
}

##################################
sub count_rows {
	my $database = shift;
	my $table = shift;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$database","","");
	return $dbh->selectrow_array("SELECT count(*) FROM $table");
}

##################################
sub query {
	my $query = shift;
	my @args = @_;
	
	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","");
	my $sth = $dbh->prepare($query);
	$sth->execute(@args);
	my @result;
	while (my @values = $sth->fetchrow_array()) {
		push @result, @values;
	}
	return @result;
}


##################################
# return longest length for one item in one column in one table (in one db, in one...)
# or 0 if no item found
sub longest {
	my $database = shift;
	my $table = shift;
	my $column = shift;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$database","","");

	$column = &get_real_column($table, $column);

	return $dbh->selectrow_array("SELECT IFNULL(length($column),0) FROM $table ORDER BY length($column) DESC LIMIT 1");
}

##################################
sub save_history {
	my @history = @_;

	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","",{AutoCommit => 0});
	my $sth = $dbh->prepare("INSERT INTO history VALUES (?)");
	for (@history) {
		$sth->execute($_);
	}
	$sth->finish();
	$dbh->commit();
}

##################################
sub init_vars {
    &db::query("CREATE TABLE IF NOT EXISTS variables (variable, value)");
    return if &db::query("SELECT * FROM variables");
    for (qw(columns uploader crawler_depth document_root debug max_sendable max_subqueries processes sleep_after_hit cookie database target::database target::user target::version)) {
        &db::query("INSERT INTO variables VALUES (?,NULL)", $_);
    }
}

##################################
sub load_vars {
    if (&db::query("SELECT value FROM variables WHERE variable = 'debug' AND value is NULL")) {
        print STDERR "[+] session \"$main::server\" created\n";
        return;
    }

    @conf::columns              = split ',', join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'columns'", $conf::columns);
    $conf::uploader             = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'uploader'", $conf::uploader);
    $conf::crawler_depth        = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'crawler_depth'", $conf::crawler_depth);
    $conf::document_root        = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'document_root'", $conf::document_root);
    $conf::debug                = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'debug'", $conf::debug);
    $conf::max_sendable         = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'max_sendable'", $conf::max_sendable);
    $conf::max_subqueries       = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'max_subqueries'", $conf::max_subqueries);
    $conf::processes       		= join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'processes'", $conf::processes);
    $conf::sleep_after_hit   = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'sleep_after_hit'", $conf::sleep_after_hit);
    $conf::cookie               = join '', &db::query("SELECT IFNULL(value,?) FROM variables WHERE variable = 'cookie'", $conf::cookie);

    $main::database             = join '', &db::query("SELECT IFNULL(value,'') FROM variables WHERE variable = 'database'");
    $main::target{database}     = join '', &db::query("SELECT IFNULL(value,'') FROM variables WHERE variable = 'target::database'");
    $main::target{user}         = join '', &db::query("SELECT IFNULL(value,'') FROM variables WHERE variable = 'target::user'");
    $main::target{version}      = join '', &db::query("SELECT IFNULL(value,'') FROM variables WHERE variable = 'target::version'");

    print STDERR "[+] session successfully loaded\n";
}

##################################
sub save_vars {
	my $dbh = DBI->connect("dbi:SQLite:dbname=$main::logdb","","",{AutoCommit => 0});
	my $sth = $dbh->prepare("UPDATE variables SET value = ? WHERE variable = ?");

	$sth->execute(join (',', @conf::columns), "columns");
	$sth->execute($conf::uploader, "uploader");
	$sth->execute($conf::crawler_depth, "crawler_depth");
	$sth->execute($conf::document_root, "document_root");
	$sth->execute($conf::debug, "debug");
	$sth->execute($conf::max_sendable, "max_sendable");
	$sth->execute($conf::max_subqueries, "max_subqueries");
	$sth->execute($conf::processes, "processes");
	$sth->execute($conf::sleep_after_hit, "sleep_after_hit");
	$sth->execute($conf::cookie, "cookie");

	$sth->execute($main::database, "database");
	$sth->execute($main::target{database}, "target::database");
	$sth->execute($main::target{user}, "target::user");
	$sth->execute($main::target{version}, "target::version");

	$sth->finish;
	$dbh->commit;
    print STDERR "[+] session successfully saved\n";
}

1
